;	MMC64 plugin for flashing Retro Replay
;
;	accepts 4/8 KB images @8000 and 32/64 KB Action Replay / Retro Replay ROMs

	processor	6502

	MAC	inc16
	inc	[{1}]
	bne	*+4		; zp only!
	inc	[{1}+1]
	ENDM

	MAC	dec16y
	ldy	[{1}]
	bne	*+4
	dec	[{1}+1]
	dec	[{1}]
	ENDM

	MAC	ldxyi
	ldx	#<[{1}]
	ldy	#>[{1}]
	ENDM

	MAC	ldxy
	ldx	[{1}]
	ldy	[{1}]+1
	ENDM

	MAC	stxy
	stx	[{1}]
	sty	[{1}]+1
	ENDM

	MAC	rtrigger
	lda	#$40
	ora	$df11
	sta	$df11
	bit	$df10
	ENDM

	MAC	wtrigger
	bit	$df10
	lda	#~$40
	and	$df11
	sta	$df11
	ENDM

	MAC	select
	lda	#~2
	and	$df11
	sta	$df11
	ENDM

	MAC	deselect
	lda	#2
	ora	$df11
	sta	$df11
	ENDM

;	For reference: memory map when changing bit 0,1&5 of $de00

; Flash jumper unset

;	8000		A000		E000		RAM	/E	/G
;							>$1000
; 00	RR ROM		BASIC		KERNAL		RAM	0	1
; 01	--		RR ROM		KERNAL		RAM	0	0
; 02	C64 RAM		BASIC		KERNAL		RAM	1	1
; 03	RR ROM		--		RR ROM		--	1	0	Ultimax
;
; 20	RR RAM
; 21	RR RAM
; 22	C64 RAM
; 23	RR RAM

; Flash jumper set
;
; 00	RR ROM		BASIC		KERNAL		RAM
; 01	--		BASIC		KERNAL		RAM
; 02	C64 RAM		BASIC		KERNAL		RAM
; 03	RR ROM		BASIC		KERNAL		RAM

; 20	RR RAM
; 21	RR RAM
; 22	C64 RAM
; 23	RR RAM

;	$de00 values

;	need to use Ultimax mode for RR_RAMRR to avoid overwriting C64 RAM
;	need to use Ultimax mode for RR_ROMRR for flashing

RR_RAM64	= $02			; /EXROM=1, /GAME=1
RR_RAMRR	= $23			; /EXROM=1, /GAME=0, RRRAM=1	(Ultimax mode, does
RR_ROMRR	= $03			; /EXROM=1, /GAME=0, RRRAM=0	not change C64 RAM)

RAM	= $20
REU	= $40

BANK0	= $00
BANK1	= $08
BANK2	= $10
BANK3	= $18
;		ROM only from this point on
BANK4	= $80
BANK5	= $88
BANK6	= $90
BANK7	= $98
;		need to use $de01 for these + flash jumper must be set
BANK8	= $20
BANK9	= $28
BANKA	= $30
BANKB	= $38
BANKC	= $a0
BANKD	= $a8
BANKE	= $b0
BANKF	= $b8


crclo		= $0800
crchi		= $0900
FATbuf		= $7800		; two pages
ZPbuf		= $7f00


	seg.u	zp
	org	2
				; FAT16 geometry
FATbase		ds.b	3	; FAT start sector
dataBase	ds.b	3	; data area start sector
clusterSize	ds.b	1	; #sectors per cluster
clusterShift	ds.b	1	; << needed to get above
startClust	ds.w	1	; first cluster of file
FATSecInMem	ds.b	1	; FAT sector in memory

size		ds.b	3	; size in bytes
fileszidx	ds.b	1	; size as index [32,64,,8,4]
secCount	ds.b	1	; size in 512 byte sectors


cluster		ds.w	1	; current cluster
sec		ds.b	3	; current sector
secLeft		ds.b	1	; this many sectors left in this cluster

cmd		ds.b	6	; backwards, [crc ll lh hl hh cmd]

src		ds.w	1
dest		ds.w	1
len		ds.w	1

tmp		ds.b	4

romszidx	ds.b	1
SZ_4		= 0
SZ_8		= 1
SZ_16		= 2
SZ_32		= 3
SZ_64		= 4

erasemask	ds.b	1	; bit=1 if blank/erased
blankmask	ds.b	1	; bit=1 if erased
bank		ds.b	1

erased_l	ds.b	1
erased_h	ds.b	1
blank_l		ds.b	1
blank_h		ds.b	1

crc		ds.w	1	; crc-16 of single bank
				; later used as chksum for full ROM image
chksum		= crc

fManufacturer	ds.b	1	; Flash manufacturer
fDevice		ds.b	1	; device
fProtection	ds.b	1	; protected sectors


scr		ds.w	1
row		ds.b	1


crctable	= $0340		; 24*CRC16


MMCBios_FirstDataSector		= $cf80
MMCBios_FATStartSector		= $cf83
MMCBios_SecPerCluster		= $cf86
MMCBios_FileSize		= $cf87
MMCBios_ClusLo			= $cf8b


;	all code run in Ultimax mode needs to be under $1000

	seg	code
	org	$0a00

	jmp	START

;	------------------------------------------------------------------

;	Flash specific

	subroutine
f_Reset
	lda	#$f0			; read/reset
f_CmdC64RAM
	clc				; return to C64 RAM afterwards
f_Command
	ldx	#RR_ROMRR
	stx	$de00
	ldx	#$50			; 5555,aa
	stx	$de01
	ldx	#$aa
	stx	$8000+$1555
	ldx	#$48			; 2aaa,55
	stx	$de01
	ldx	#$55
	stx	$8000+$0aaa
	ldx	#$50			; 5555,<cmd>
	stx	$de01
	sta	$8000+$1555

;	min 5 us delay if done during erase/program instruction
;	luckily we are running on a slow computer :)

	bcs	.x			; Fc=1? keep ROM
	ldx	#RR_RAM64		; C64 RAM
	stx	$de00
.x	rts

;	----

;	in:	xy=address, a=data
;	out:	Fc=0 for success
;		Fc=1 if Flash timeout

	subroutine
f_Poll
	stxy	dest
f_Poll2
	and	#$80
	sta	tmp

	ldy	#0
.1	lda	(dest),y
	eor	tmp			; b7(flash) = b7(in)? success
	bpl	.x

	and	#$20			; test b5 for timeout
	beq	.1

	lda	(dest),y		; re-test b7
	eor	tmp

.x	asl				; Fc=1 if error
	rts

;	----

	subroutine
f_AutoSelect
	sec				; keep ROM
	lda	#$90
	jsr	f_Command

	lda	$8000
	sta	fManufacturer
	lda	$8001
	sta	fDevice

	ldx	#2*7
.3	lda	de01_ROM,x
	sta	$de01
	lda	$8002
	lsr
	rol	fProtection
	dex
	dex
	bpl	.3

	jsr	f_Reset
	rts

;	---

;	in:	A=bitmask of sectors to erase
;	erase in two chunks to ensure <80 us timing

	subroutine
f_Erase
	pha
	ldy	#0			; clear sectors in lower 64 KB
	jsr	.1
	pla
	bcs	.x
	ldy	#8			; clear sectors in upper 64 KB
	lsr
	lsr
	lsr
	lsr
	jsr	.1
.x	rts

.1	clc
	and	#$0f
	beq	.x

	sta	tmp

	sec				; keep ROM
	lda	#$80
	jsr	f_Command
	lda	#$aa			; $5555,$aa
	sta	$9555
	lda	#$48
	sta	$de01
	lda	#$55			; $2aaa,$55
	sta	$8aaa

;	this loop must assure <80 cycles between $8000 writes
;	so sectors can be erased simultaneously
;
;	worst bit combination 1000 has 17+14+14+18=63 cycle delay

	lda	tmp		;  3
.2	lsr			;  2	 2
	bcc	.3		;  4	 5
	ldx	de01_ROM,y	;  8
	stx	$de01		; 12
	ldx	#$30		; 14
	stx	$8000		; 18
.3	iny			; 20	 7
	iny			; 22	 9
	cpy	#16		; 24	11
	bcc	.2		; 27	14

;	80 us delay to start erasing

.4	dey				; 16*5+4=84
	bpl	.4

	lda	#8			; wait until erase starts
.5	bit	$8000
	beq	.5

	ldxyi	$8000			; wait for completion
	lda	#$ff
	jsr	f_Poll

	php
	jsr	f_Reset
	plp
	rts

;	----

;	(src)	points to memory source
;	(dest)	points to flash destination
;	len.w	is #bytes to program
;	bank	is $de01 value for flash bank
	subroutine
f_Program

	jmp	.3

.1	sec
	lda	#$a0
	jsr	f_CmdC64RAM

	ldy	#0
	lda	(src),y
	ldx	#RR_ROMRR		; RR ROM
	stx	$de00
	ldx	bank			; RR flashROM
	stx	$de01
	sta	(dest),y

	inc16	src			;  8 (+4)
	dec16y	len			; 11 (+4)

	jsr	f_Poll2
	bcs	.4

	inc16	dest			;  8 (+4)

.3	lda	len
	ora	len+1
	bne	.1

	clc
.4	php
	jsr	f_Reset
	plp
	rts

;	----

	subroutine
FindRR
	lda	#RR_RAM64		; write $55 to C64 RAM
	sta	$de00
	lda	#$55
	sta	$8000

	lda	#RR_RAMRR		; write 03/0c/30/c0 to RR RAM
	sta	$de00

	lda	#$03
	sta	tmp
	lda	#RAM|BANK0|REU
	clc
.1	sta	$de01
	adc	#8
	ldx	tmp
	stx	$8000
	asl	tmp
	asl	tmp
	bcc	.1

	lda	#RR_RAM64		; check C64 RAM
	sta	$de00
	lda	#$55
	sta	tmp
	cmp	$8000
	bne	.3			; overwritten

	lda	#RR_RAMRR		; check RR RAM
	sta	$de00

	lda	#$03
	sta	tmp
	lda	#RAM|BANK0|REU
	clc
.2	sta	$de01
	adc	#8
	ldx	tmp
	cpx	$8000
	bne	.3			; overwritten
	asl	tmp
	asl	tmp
	bcc	.2

.3	lda	#RR_RAM64
	sta	$de00

	ldy	#5
	jsr	SetY
	ldxyi	RR_txt
	jsr	Print

	ldxyi	RR_ok
	lda	tmp
	beq	.4
	ldxyi	RR_err
.4	jsr	Print

	lda	tmp
	cmp	#1			; Fc=1 if >0
	rts

;	----

	subroutine
RamTest

	ldy	#11
	jsr	SetY
	ldxyi	RAM_txt
	jsr	Print

;	fill memory with pattern

	lda	#RR_RAMRR
	sta	$de00

	ldx	#$7f
	txa
.f1					; x=a=7f/5f/3f/1f
	ldy	#$9f
	sty	src+1
	ldy	#$00
	sty	src

	lsr
	lsr
	lsr
	lsr
	lsr				; 3-0
	eor	#$34			; "7"-"4"
	sta	(scr),y			; show onscreen

	txa
	lsr
	lsr				; 000??111
	and	#BANK3			; 000??000	00/08/10/18
	ora	#REU			; 010??000
	sta	$de01

.f2	txa
.f3	sta	(src),y
	clc
	adc	#1
	iny
	bne	.f3

	dex
	dec	src+1			; do until 8 KB
	bmi	.f2

	txa
	bpl	.f1			; do until 32 KB

;	read it back

	ldx	#$7f
	txa
.c1
	ldy	#$9f
	sty	src+1
	ldy	#$00

	lsr
	ror
	ror
	lsr
	lsr
;	ora	#$30			; "3"-"0"
	sta	(scr),y

	txa
	lsr
	lsr
	and	#BANK3
	ora	#REU
	sta	$de01

.c2	txa
.c3	cmp	(src),y
	bne	.fail
;	sec
	adc	#1-1
	iny
	bne	.c3

	dex
	dec	src+1			; do until 8 KB
	bmi	.c2
	txa
	bpl	.c1			; do until 32 KB

	clc
	dc.b	$24
.fail	sec
	php

	lda	#RR_RAM64
	sta	$de00

	ldxyi	OK_txt
	bcc	.rt_ok
	ldxyi	ERROR_txt
.rt_ok	jsr	Print

	plp
	rts

;	----

	subroutine
RomTest

	ldy	#12
	jsr	SetY
	ldxyi	LBank_txt
	jsr	Print

	lda	#0

.1	sta	tmp
	tax

	ldy	#0
	and	#7
	eor	#$37
	sta	(scr),y

	lda	#RR_ROMRR
	sta	$de00
	lda	de01_ROM,x
	sta	$de01
	sta	bank
	jsr	ChkSum
	lda	#RR_RAM64
	sta	$de00

	lda	tmp
	cmp	#7
	bne	.4

	lda	erased_h
	sta	erased_l
	lda	blank_h
	sta	blank_l

	ldy	#0			; calculate ROM size,
	ldx	#erased_l		; show checksum
	jsr	ROMsize

	ldy	#13			; show ROM name
	ldx	#12
	jsr	SetXY
	ldy	#0
	jsr	ShowROMType

	ldy	#14
	jsr	SetY
	ldxyi	UBank_txt
	jsr	Print

.4	inc	tmp
	lda	tmp
	cmp	#16
	bne	.1

	ldy	#16
	ldx	#erased_h
	jsr	ROMsize

	ldy	#15			; show ROM name
	ldx	#12
	jsr	SetXY
	ldy	#16
	jsr	ShowROMType


	lda	#0			; check for bank jumper unset
	sta	tmp
	ldx	#14
.5	lda	crctable,x
	eor	crctable+16,x
	ora	tmp
	sta	tmp
	lda	crctable+1,x
	eor	crctable+17,x
	ora	tmp
	sta	tmp
	dex
	dex
	bpl	.5
	tax
	bne	.6

	ldy	#16
	jsr	SetY
	ldxyi	BJumper_txt
	jsr	Print

.6

	lda	erased_h		; combine 8KB bank flags into 16 KB sector flags
	jsr	.sub
	lda	erased_l
	jsr	.sub
	lda	tmp
	sta	erasemask		; bit=1 if 8KB erased/blank

	lda	blank_h
	jsr	.sub
	lda	blank_l
	jsr	.sub
	lda	tmp
	sta	blankmask		; bit=1 if 8KB erased, 0 otherwise
	rts

.sub	jsr	.s1
.s1	jsr	.s2
.s2	cmp	#$c0			; Fc=1 if 16KB erased
	rol	tmp
	asl
	asl
	rts

;	----

	subroutine
CrcPage
.1	lda	(src),y
	eor 	crc+1
	tax
	lda 	crc
	eor 	crchi,x
	sta 	crc+1
	lda 	crclo,x
	sta 	crc
	iny
	bne	.1
	rts


;	calculate CRC-16 of 8KB bank

	subroutine
ChkSum

	jsr	PatchIfNeeded

	ldy	#$80
	sty	src+1
	ldy	#0
	sty	src

	lda	#-1
	sta	crc
	sta	crc+1
	lda	#32
	sta	tmp+1

.1	jsr	CrcPage
	inc	src+1
	dec	tmp+1
	bne	.1

	dec	src+1			; blank/erased?
	lda	(src),y
	beq	.blank
	cmp	#$ff
	beq	.erased

.2	lsr	erased_h
	lsr	blank_h
	bpl	.4

.blank
.erased

.3	cmp	(src),y
	bne	.2			; not 00/ff
	iny
	bne	.3
	dec	src+1
	bmi	.3

	lsr
	ror	erased_h		; bit=1 if erased
	eor	#1
	lsr
	ror	blank_h			; bit=1 if blank

.4	lda	tmp			; save crc-16
	asl
	tax
	lda	crc
	sta	crctable,x
	lda	crc+1
	sta	crctable+1,x
	rts

;	----

;	if bank needs patching, copy it to C64 RAM and do it
;	returns data at $8000 in any case

	subroutine
PatchIfNeeded
	ldy	#0
	sty	tmp+1
	sty	src
	sty	tmp+2

.1	ldx	tmp+1
	lda	PatchPage,x		; page to check
	beq	.x

	cmp	tmp+2			; have we checksum for this page?
	beq	.3
	sta	tmp+2

	sta	src+1
	dey
	sty	crc
	sty	crc+1
	iny
	jsr	CrcPage

.3	lda	tmp+1
	asl
	tax
	lda	crc
	cmp	PatchCrc,x
	bne	.next
	lda	crc+1
	cmp	PatchCrc+1,x
	beq	.patch			; found, copy to RAM & patch

.next	inc	tmp+1
	bne	.1

.x	rts


.patch	lda	bank			; skip if already in C64 RAM
	beq	.p0

	lda	#$9f			; copy to RAM
	sta	.c2+2
	sta	.c4+2

.c1	lda	#RR_ROMRR
	sta	$de00
	lda	bank
	sta	$de01

.c2	lda	$8000,y			; need to copy through low RAM
	sta	$0200,y
	iny
	bne	.c2

	lda	#RR_RAM64
	sta	$de00

.c3	lda	$0200,y
.c4	sta	$8000,y
	iny
	bne	.c3

	dec	.c2+2
	dec	.c4+2
	bmi	.c1			; continue until 8KB copied

;	not in Ultimax mode anymore, patch data can be anywhere

.p0	lda	PatchPtr,x
	sta	src
	lda	PatchPtr+1,x
	sta	src+1

.p1	ldy	#0
	lda	(src),y
	beq	.x
	sta	dest+1
	iny
	lda	(src),y
	sta	dest
	iny
	lda	(src),y
	tay

	clc
	lda	#3
	adc	src
	sta	src
	bcc	.p2
	inc	src+1

.p2	lda	#$ff
.p3	sta	(dest),y
	dey
	bpl	.p3
	bmi	.p1

;	----

	subroutine

;	determine ROM size, calculate chksum

.blank	ldxyi	Blank_txt
	lda	#0
	beq	.x1
.erased	ldxyi	Erased_txt
.x1	sta	chksum
	sta	chksum+1
	jsr	Print
	rts

ROMsize
	sty	tmp+1

	lda	#$ff
	cmp	0,x
	beq	.erased
	cmp	2,x
	beq	.blank

	lda	#SZ_64
	sta	romszidx

	ldy	#8
	ldx	tmp+1
.1	lda	crctable,x
	cmp	crctable+8,x
	bne	.64
	inx
	dey
	bne	.1

;	bank[0-3] == bank[4-7]

	dec	romszidx

	ldx	tmp+1
	ldy	#4
.2	lda	crctable,x
	cmp	crctable+4,x
	bne	.32
	inx
	dey
	bne	.2

;	bank[0-1] == bank[2-3]

	dec	romszidx

	ldx	tmp+1
	lda	crctable,x
	cmp	crctable+2,x
	bne	.16
	lda	crctable,x
	cmp	crctable+2,x
	bne	.16

;	bank[0] == bank[1], either 4 KB or 8 KB

	dec	romszidx

	cpx	#$10			; 00=lower,10=upper,20=ram
	beq	.u
	bcc	.l
	lda	#RR_RAM64
	sta	$de00
	bne	.7
.l	lda	#BANK0|REU
	dc.b	$2c
.u	lda	#BANK8|REU

	ldy	#RR_ROMRR
	sty	$de00

	sta	$de01

.7	ldy	#0
	sty	src
	sty	dest
	lda	#$80
	sta	src+1
	lda	#$90
	sta	dest+1
	ldx	#16
.9	lda	(src),y
	cmp	(dest),y
	bne	.8
	iny
	bne	.9
	inc	src+1
	inc	dest+1
	dex
	bne	.9

.4	dec	romszidx
.8	lda	#RR_RAM64
	sta	$de00

.16
.32
.64

;	combine partial CRCs into single checksum

	lda	#0
	sta	chksum
	sta	chksum+1

	lda	romszidx
	cmp	#SZ_16
	bcc	.chk48
	beq	.chk16
	cmp	#SZ_32
	beq	.chk32

.chk64	ldy	#8
	dc.b	$2c
.chk32	ldy	#4
	dc.b	$2c
.chk16	ldy	#2
	dc.b	$2c
.chk48	ldy	#1

	ldx	tmp+1
.10	lda	chksum+1
	asl
	rol	chksum
	rol	chksum+1

	lda	crctable,x
	eor	chksum
	sta	chksum
	lda	crctable+1,x
	eor	chksum+1
	sta	chksum+1
	inx
	inx
	dey
	bne	.10

	ldx	romszidx
	lda	RomSzBCD,x
	jsr	PrHex
	jsr	NoZero
	ldxyi	kb_txt
	jsr	Print
	lda	chksum+1
	jsr	PrHex
	dec16y	scr
	lda	chksum
	jsr	PrHex
	rts

;	------------------------------------------------------------------

;	Configurable ROMs mess up the CRC calculation, so we fill changing
;	bytes with $ff. To do that we need to check if current bank is one
;	of these:

;ROM		0000	2000	6000

;tfr01			b9da	b2ed		3e00:10, 7e00:10
;tfr02			edbc	b2ed		3e00:10, 7e00:10
;tfr03		3dd2		80a7		1000:14
;tfr04		dc65		d24f		1000:14
;tfr05		f1b1		d24f		1000:14
;tfr06		f1b1		d24f		1000:14
;
;hl040707	c675				21c3:10
;hl040714	c675				21c3:10
;hl040919	85fc				2263:10

;	page to check

PatchPage	HEX	8080			; TFR 0.1	2000/6000
		HEX	80			; TFR 0.2	2000		6000 = 0.1
		HEX	8080			; TFR 0.3	0000/6000
		HEX	8080			; TFR 0.4	0000/6000
		HEX	80			; TFR 0.5/0.6	0000		6000 = 0.4

		HEX	80			; HL 07		0000
		HEX	80			; HL 09		0000
		HEX	00

;	CRC-16 of page

PatchCrc	dc.w	$b9da,$b2ed
		dc.w	$edbc
		dc.w	$3dd2,$80a7
		dc.w	$dc65,$d24f
		dc.w	$f1b1

		dc.w	$c67c
		dc.w	$85fc

;	----

	subroutine

ShowROMType
	sty	tmp+1

	ldxyi	KnownROMs
	stxy	src
.5	ldy	#0
	lda	(src),y
	sta	tmp+3
	bmi	.unknown

	iny
	and	#7			; check ROM size and CRC
	cmp	romszidx
	bne	.skip1
	lda	(src),y
	iny
	cmp	chksum
	bne	.skip
	lda	(src),y
	cmp	chksum+1
	bne	.skip
	beq	.found

.skip1	iny				; skip name
.skip	iny
	lda	(src),y
	bpl	.skip

	jsr	.sub
	jmp	.5

.unknown
.found	jsr	.sub
	jmp	Print2

.sub	sec
	tya
	adc	src
	sta	src
	bcc	.sx
	inc	src+1
.sx	rts

;	----

	subroutine
START
	sei
	jsr	Init

	lda	#RR_RAM64
	sta	$de00

	jsr	Cls

	ldx	#4
	ldy	#0
	jsr	SetXY
	ldxyi	Title_txt
	jsr	Print

	jsr	IdentFile		; identify input file

Retry
	ldx	#0			; what a cheap bastard I am...
	lda	#$20
.r	sta	$0400+4*40,x
	sta	$0500,x
	sta	$0600,x
	sta	$0700,x
	inx
	bne	.r

;	init RR hardware

	jsr	FindRR
	bcs	Wait			; fatal error

;	we have *something* with 32 KB RAM,
;	check for flashmode jumper

	ldy	#6
	jsr	SetY
	ldxyi	FJumper_txt
	jsr	Print

	lda	$de01
	ldxyi	FJumper_ok
	lsr
	php
	bcs	.j_ok
	ldxyi	FJumper_err
.j_ok	jsr	Print
	plp
	bcc	Wait			; fatal error

	jsr	IdentChip		; check flash chip

	jsr	RamTest			; do full RAM check
	jsr	RomTest			; try to identify current ROMs


Wait
;	continue/rescan

	ldy	#18
	jsr	SetY
	ldxyi	Exit_txt
	jsr	Print

	lda	$0400+2*40+12		; "Input file: [b]ad size"
	cmp	#$02
	beq	.keys
	lda	#$0e			; 'n'
	cmp	$0400+6*40+17		; "Flashmode jumper [n]ot set"
	beq	.retry
	cmp	$0400+9*40+19		; "Protected sectors: [n]one"
	bne	.keys

	ldy	#21
	jsr	SetY
	ldxyi	Bank_txt
	jsr	Print

	lda	#$0e
	lda	$0400+16*40		; "[W]arning: bank-select jumper may be open"
	cmp	#$57
	bne	.keys

.retry	ldy	#19
	jsr	SetY
	ldxyi	Retry_txt
	jsr	Print

;	run/stop	7f 80
;	R		fb 02
;	U		f7 40
;	L		df 04

.keys	jsr	RdKeys
	bmi	.k0

	lda	$de01			; don't exit if FJ set
	lsr
	bcs	.k0
	jmp	Exit

.k0	lda	$0400+19*40		; "[P]ress R to retry after checking jumpers"
	cmp	#$50
	bne	.k1

	txa
	and	#$02
	bne	.k1
	jmp	Retry

.k1	lda	$0400+21*40		; "[U]pper or lower bank?"
	cmp	#$55
	bne	.keys

	txa
	asl
	bpl	.upper
	and	#4<<1
	bne	.keys

.lower	lda	#"L"
	dc.b	$2c
.upper	lda	#"U"
	sta	$0400+21*40+21
	eor	#"L"
	beq	.bl
	lda	#-1
.bl	sta	bank

	ldx	#0
.cpy	lda	$0400+21*40,x
	sta	$0400+18*40,x
	inx
	cpx	#4*40
	bcc	.cpy

	jsr	Erase
	bcs	Wait2
	jsr	Flash

Wait2	php

	ldx	#27
	lda	#$20
.crom	sta	$0400+12*40+12,x
	sta	$0400+13*40+12,x
	sta	$0400+14*40+12,x
	sta	$0400+15*40+12,x
	dex
	bpl	.crom

	jsr	RomTest			; re-ident ROMs
	plp
	bcc	.w2a

	lda	#$ff
	sta	retry_mod

	ldy	#22
	jsr	SetY
	ldxyi	Retry_txt
	jsr	Print

	lda	#$20
	sta	retry_mod

.w2a	ldy	#23
	jsr	SetY
	ldxyi	PowerOff_txt
	jsr	Print
	lda	#$70
	sta	Exit_txt
	ldy	#24
	jsr	SetY
	ldxyi	Exit_txt
	jsr	Print
	lda	#$50
	sta	Exit_txt

.keys2	jsr	RdKeys
	bmi	.k20

	lda	$de01			; don't exit if FJ set
	lsr
	bcc	Exit

.k20	lda	$0400+22*40		; "[P]ress R to retry"
	cmp	#$50
	bne	.keys2

	txa
	and	#2
	bne	.keys2
	jmp	Retry

Exit	jsr	f_Reset
	lda	#$66			; enable BIOS, disable RR
	sta	$df11
	jsr	RestoreZP
	bit	$dc0d			; ack timer interrupts
	bit	$dd0d
	rts

;	----

	subroutine
RdKeys
	lda	#0
	sta	tmp
	ldx	#3
.1	lda	.dc00,x
	sta	$dc00
	lda	.dc01,x
	and	$dc01
	ora	tmp
	sta	tmp
	dex
	bpl	.1
	tax
	rts

;		rs R  U  L

.dc00	HEX	7f fb f7 df
.dc01	HEX	80 02 40 04


;	------------------------------------------------------------------

	subroutine
IdentFile
	ldy	#2
	jsr	SetY
	ldxyi	File_txt
	jsr	Print

	lda	size+2
	lsr
	ora	MMCBios_FileSize+3
	ora	size
	bne	.sz_err

	lda	size+1
	ror					; 08/10/40/80 for 4/8/32/64 KB
	bcs	.sz_err
	beq	.sz_err

;	76543210	bit
;
;	6310000.	ROM
;	42684215	size
;
;	43210...	szidx

	ldx	#-4
.sz_1	inx
	lsr
	bcc	.sz_1
	bne	.sz_err
	cpx	#5
	bcs	.sz_err
	cpx	#2
	bne	.sz_ok

.sz_err	ldxyi	Size_err
	jsr	Print
	sec
	rts

.sz_ok	stx	fileszidx			; 0/1/2/3/4
	lda	RomSecSz,x
	sta	secCount

;	checksum whole image

	pha

	lda	startClust
	sta	cluster
	lda	startClust+1
	sta	cluster+1
	jsr	StartCluster

	ldx	#$10
.1	stx	tmp

	ldy	#0
	lda	secCount
	lsr
	lsr
	lsr
	lsr
	sbc	#1-1
	eor	#$30
	sta	(scr),y

	jsr	ChkSum8KB

	ldx	tmp
	inx
	cpx	#$18
	bne	.1

	pla
	sta	secCount

	ldy	#$20
	ldx	#erased_h
	jsr	ROMsize

	ldy	#3			; show image name
	ldx	#12
	jsr	SetXY
	ldx	#erased_h
	ldy	#32
	jsr	ShowROMType

	clc
	rts


	subroutine

;	x=[$10,$17]

ChkSum8KB
	lda	secCount
	beq	.1

	jsr	Read8KB
	lda	#0
	sta	bank
	jsr	ChkSum
	rts

;	fake 64 KB rom by reusing earlier crc-16

.1	lda	.fake-16,x
	tay
	txa
	asl
	tax

	lda	crctable,y
	sta	crctable,x
	lda	crctable+1,y
	sta	crctable+1,x

	lsr	erased_h
	lsr	blank_h
	rts

.fake	HEX	20 20 20 22 20 22 24 26

;	----

	subroutine
Read8KB
	ldxyi	$8000
	stxy	dest

.1	jsr	ReadSector
	jsr	NextSector
	inc	dest+1
	inc	dest+1
	dec	secCount

	lda	dest+1
	and	#$1f
	beq	.x			; 8 KB loaded
	lda	secCount
	bne	.1			; no EOF

;	has to be 4 KB ROM, copy to upper half too

	tay
	sty	src
	sec
	lda	dest+1
	sbc	#$10
	sta	src+1
	ldx	#$10
.2	lda	(src),y
	sta	(dest),y
	iny
	bne	.2
	inc	src+1
	inc	dest+1
	dex
	bne	.2

.x	rts

;	----

	subroutine
IdentChip

	jsr	f_AutoSelect		; read manufacturer, device & protection status

	ldy	#7
	jsr	SetY
	ldxyi	ID_txt
	jsr	Print

	lda	fManufacturer
	jsr	PrHex
	lda	fDevice
	jsr	PrHex

	ldy	#8
	jsr	SetY

	lda	#0
.1	tax
	lda	fManufacturer
	cmp	.chip,x
	bne	.2
	lda	fDevice
	cmp	.chip+1,x
	beq	.found
.2	clc
	txa
	adc	#6
	cmp	#4*6
	bcc	.1

	ldxyi	Unknown_txt
	jsr	Print

.err	sec
	rts

.found	lda	.chip+5,x
	pha
	lda	.chip+4,x
	pha
	ldy	.chip+3,x
	lda	.chip+2,x
	tax
	jsr	Print
	pla
	tax
	pla
	tay
	jsr	Print

;	show protected sectors

	ldy	#9
	jsr	SetY
	ldxyi	Prot_txt
	jsr	Print

	lda	fProtection
	beq	.p_ok
	jsr	PrHex
	clc
	rts

.p_ok	ldxyi	Prot_ok
	jsr	Print
	clc
	rts

;	01 20	AMD 29F010
;	01 A4	AMD 29F040
;	20 E2	ST  29F040

.chip	dc.w	$2001,AMD_txt,d29F010_txt
	dc.w	$a401,AMD_txt,d29F040_txt
	dc.w	$2020,ST_txt,d29F010_txt
	dc.w	$e220,ST_txt,d29F040_txt

;	----

	subroutine
Erase

	ldy	#19
	jsr	SetY
	ldxyi	Erase_txt
	jsr	Print

	lda	erasemask
	and	blankmask
	eor	#$ff
	sta	erasemask

	lda	bank
	and	#$f0
	bmi	.1
	lda	#$0f
.1	and	erasemask
	bne	.2

	ldxyi	Skipped_txt
	jsr	Print
	clc
	rts

.2	jsr	f_Erase

	php
	ldxyi	OK_txt
	bcc	.3
	ldxyi	ERROR_txt
.3	jsr	Print

	plp
	rts

;	----

	subroutine
Flash

	ldy	#20
	jsr	SetY
	ldxyi	Flash_txt
	jsr	Print

	lda	startClust
	sta	cluster
	lda	startClust+1
	sta	cluster+1
	jsr	StartCluster

	ldx	fileszidx
	lda	RomSecSz,x
	sta	secCount
	cpx	#SZ_32
	bcs	.big

;	4/8 KB image

;	 4 KB	same 4KB to upper half, then drop to 8KB
;	 8 KB	program 8 banks with same image

	jsr	Read8KB			; takes care of 4 KB mirroring

	lda	bank
	and	#$08
	tax

.s1	stx	tmp+1
	lda	de01_ROM,x
	sta	bank

	txa
	and	#7
	eor	#$37
	ldy	#0
	sta	(scr),y

	ldxyi	$8000
	stxy	src
	stxy	dest
	ldy	#$20
	stxy	len

	jsr	f_Program
	bcs	.err

	ldx	tmp+1
	inx
	txa
	and	#7
	bne	.s1
	beq	.ok


;	32 KB	0->0,4	1->1,5	2->2,6	3-> 3,7
;	64 KB	all different

.big	cpx	#SZ_64
	beq	.rr

.ar	lda	bank
	and	#$08
	tax

	lda	#"7"
	ldy	#0
	sta	(scr),y

.ar1	stx	tmp+1
	lda	de01_ROM,x
	sta	bank

	jsr	Read8KB

.ar2	ldxyi	$8000
	stxy	src
	stxy	dest
	ldy	#$20
	stxy	len

	jsr	f_Program
	bcs	.err

	ldy	#0
	lda	(scr),y
	sbc	#1-1
	sta	(scr),y

	lda	bank			; mirror in upper half
	eor	#$80
	sta	bank
	bmi	.ar2

	ldx	tmp+1
	inx
	txa
	and	#3
	bne	.ar1

.ok	clc
	ldxyi	OK_txt
	bne	.x
.err	ldxyi	ERROR_txt
	sec

.x	php
	jsr	Print
	plp
	rts

.rr	lda	bank
	and	#$08
	tax

	lda	#"7"
	ldy	#0
	sta	(scr),y

.rr1	stx	tmp+1
	lda	de01_ROM,x
	sta	bank

	jsr	Read8KB

	ldxyi	$8000
	stxy	src
	stxy	dest
	ldy	#$20
	stxy	len

	jsr	f_Program
	bcs	.err

	ldy	#0
	lda	(scr),y
	sbc	#1-1
	sta	(scr),y

	ldx	tmp+1
	inx
	txa
	and	#7
	bne	.rr1
	beq	.ok

	subroutine
RestoreZP
	ldx	#0
.3	lda	ZPbuf,x
	sta	0,x
	inx
	bpl	.3
	rts

	subroutine
WaitRast
.1	bit	$d011
	bpl	.1
.2	bit	$d011
	bmi	.2
	rts

	subroutine
Cls
	lda	#$0b
	sta	$d011
	jsr	WaitRast

	lda	#$0e
	sta	$d020
	lda	#6
	sta	$d021

	ldx	#0
.1	lda	#$20
	sta	$0400,x
	sta	$0500,x
	sta	$0600,x
	sta	$0700,x
	lda	#1
	sta	$d800,x
	sta	$d900,x
	sta	$da00,x
	sta	$db00,x
	inx
	bne	.1

	lda	#$1b
	sta	$d011
	jsr	WaitRast
	rts

	subroutine
SetY	ldx	#0
SetXY	txa
	ldx	#4
	stx	scr+1
	bpl	.2
.1	clc
	adc	#40
	bcc	.2
	inc	scr+1
.2	dey
	bpl	.1
	sta	scr
	rts

Print
	stxy	src
Print2
	ldy	#0
.3	lda	(src),y
	bmi	.x
	cmp	#$60			; 60..7f -> 00..1f
	bcc	.4
	and	#$1f
.4	sta	(scr),y
	iny
	bne	.3
.x	clc
	tya
	adc	scr
	sta	scr
	bcc	.x2
	inc	scr+1
.x2	rts

PrHex	ldy	#0
	pha
	lsr
	lsr
	lsr
	lsr
	jsr	PrHex4
	pla
	and	#$0f
	jsr	PrHex4
	iny
	bpl	.x

PrHex4	ora	#$30
	cmp	#$3a
	bcc	.h3
	adc	#6
.h3	sta	(scr),y
	iny
	rts

	subroutine
NoZero	ldy	#-3
	dec	scr+1
	lda	(scr),y
	cmp	#$30
	bne	.x
	lda	#$20
	sta	(scr),y
.x	inc	scr+1
	rts


	subroutine
Init

	ldx	#0
.1	lda	$00,x
	sta	ZPbuf,x
	inx
	bpl	.1
	lda	#$2f
	sta	0
	lda	#$35
	sta	1
	ldx	#2
	lda	#0
.2	sta	0,x
	inx
	bpl	.2

	lda	#$0f			; disable BIOS, enable RR
	sta	$df11

	lda	#$14			; AR (mirror of $de00)
	sta	$de02			;  disable
	lda	#$44			; RR
	sta	$de01			;  REU compatible, disable freeze
	lda	#$42			; RR
	sta	$de00			;  init freeze flip flop, C64 RAM

	ldx	#160/8
	lda	#$ff
.3	sta	$df10
	dex
	bne	.3

	ldx	#2
.4	lda	MMCBios_FATStartSector,x
	sta	FATbase,x
	lda	MMCBios_FirstDataSector,x
	sta	dataBase,x
	lda	MMCBios_FileSize,x
	sta	size,x
	dex
	bpl	.4
	stx	FATSecInMem

	lda	MMCBios_ClusLo
	sta	startClust
	lda	MMCBios_ClusLo+1
	sta	startClust+1

	lda	MMCBios_SecPerCluster
	sta	clusterSize

;	ldx	#-1			; clusterSize -> shift
.5	inx
	lsr
	bcc	.5
	stx	clusterShift

	jsr	BuildCRC
	jsr	DisableNMI
	lda	#$37
	sta	1
	rts


;	------------------------------------------------------------------
;
;	MMC64 access


	subroutine
SendRdCmd
	lda	#$40|$11	; READ_SINGLE_BLOCK
	sta	cmd+5
	lda	#0		; pos = sec * $0200
	sta	cmd+1
	lda	sec
	asl
	sta	cmd+2
	lda	sec+1
	rol
	sta	cmd+3
	lda	sec+2
	rol
	sta	cmd+4
	lda	#$ff
	sta	cmd+0

	deselect
	ldy	#$ff
	sty	$df10
	and	#~2			; select card to re-sync SPI
	sta	$df11

	ldy	#5
.1	lda	cmd,y
	sta	$df10
	dey
	bpl	.1

.2	sty	$df10
	lda	$df10
	cmp	#$ff
	beq	.2
	rts

	subroutine
WaitForData

	ldy	#0
.1	lda	#$ff
	sta	$df10
	eor	$df10
	bne	.4			; data/error token
	iny
	bne	.1
	lda	#-1
.4	cmp	#2			; Fc=1 if error
	rts



;	read single [sec] to [dest]

	subroutine
ReadSector

	jsr	SendRdCmd
	jsr	WaitForData

	rtrigger
	ldy	#0
.1	lda	$df10
	sta	(dest),y
	iny
	bne	.1
	inc	dest+1
.2	lda	$df10
	sta	(dest),y
	iny
	bne	.2
	wtrigger

	dec	dest+1

	clc
	rts


;	--------------------------------------------------------
;
;	FAT16


;	sector -> next

	subroutine
NextSector
	inc	sec
	bne	.s
	inc	sec+1
	bne	.s
	inc	sec+2
.s	dec	secLeft			; more sectors to do?
	bne	.x			; Fc=0

;	cluster -> next, set sec

NextCluster

	jsr	ClustOK
	bcs	.x

	lda	cluster+1
	jsr	ReadFATSector

	lda	cluster
	asl
	tax
	bcs	.upper
.lower	lda	FATbuf,x
	ldy	FATbuf+1,x
	bcc	.0
.upper	lda	FATbuf+256,x
	ldy	FATbuf+257,x
.0

StartClusterAY
	sta	cluster
	sty	cluster+1

;	cluster -> sector
StartCluster

	jsr	ClustOK
	bcs	.x

;	clc
	lda	cluster
	sbc	#2-1
	tay
	lda	cluster+1
	sbc	#0
	sta	sec+1
	lda	#0
	sta	sec+2

	tya
	ldx	clusterShift
	beq	.2
.1	asl
	rol	sec+1
	rol	sec+2
	dex
	bne	.1

.2	clc
	adc	dataBase
	sta	sec
	lda	sec+1
	adc	dataBase+1
	sta	sec+1
	lda	sec+2
	adc	dataBase+2
	sta	sec+2

	lda	clusterSize
	sta	secLeft

	clc
.x	rts

	subroutine

;	!! should check for ..numClusters

ClustOK
	lda	cluster
	cmp	#2
	lda	cluster+1
	sbc	#0
	bcc	.err

.1	lda	cluster
	adc	#9-1			; Fc=1 if $f7..$ff
	lda	cluster+1
	adc	#0			; Fc=1 if $fff7..$ffff
	rts
.err	sec
	rts



;	in: A=sector relative to FAT start
;	writes back modified sectors automatically

	subroutine
ReadFATSector

	cmp	FATSecInMem
	beq	.x
	sta	FATSecInMem

	clc
	adc	FATbase
	sta	sec
	lda	#0
	adc	FATbase+1
	sta	sec+1
	lda	#0
	adc	FATbase+2
	sta	sec+2

	lda	dest
	pha
	lda	dest+1
	pha

	lda	#<FATbuf
	ldy	#>FATbuf
	sta	dest
	sty	dest+1
	jsr	ReadSector

	pla
	sta	dest+1
	pla
	sta	dest

.x	clc
	rts

;	------------------------------------------------------------------

	subroutine
BuildCRC
	ldx	#0
.1	lda	#0			; low 8 bits of the crc-16
	stx	crc+1			; high 8 bits

	ldy	#8			; 8 bits
.2	asl
	rol	crc+1
	bcc	.3

	eor	#$21			; add crc-16 polynomial $1021
	pha
	lda	crc+1
	eor	#$10
	sta	crc+1
	pla

.3	dey
	bne	.2

	sta	crclo,x
	lda	crc+1
	sta	crchi,x
	inx
	bne	.1
	rts

;	------------------------------------------------------------------

	subroutine
DisableNMI
	lda	#<.rti
	sta	$fffa
	lda	#>.rti
	sta	$fffb
	lda	#$81
	sta	$dd0d			; CIA2 TA IR enable
	ldx	#0			; IRQ in one cycle
	stx	$dd05
	inx
	stx	$dd04
	lda	#$dd
	sta	$dd0e			; cause it now
	rts
.rti	rti

Title_txt	dc.b	"Retro Replay Flasher plugin V1.0",255
OK_txt		dc.b	"ok",255
ERROR_txt	dc.b	"error",255

File_txt	dc.b	"Input file: ",255
Size_err	dc.b	"bad size",255

RR_txt		dc.b	"Retro Replay RAM ",255
RR_err		dc.b	"not "
RR_ok		dc.b	"found",255

ID_txt		dc.b	"Flash manufacturer & device: ",255
AMD_txt		dc.b	"AMD ",255
ST_txt		dc.b	"ST ",255
d29F010_txt	dc.b	"29F010",255
d29F040_txt	dc.b	"29F040",255

FJumper_txt	dc.b	"Flashmode jumper ",255
FJumper_err	dc.b	"not "
FJumper_ok	dc.b	"set",255

Prot_txt	dc.b	"Protected sectors: ",255
Prot_ok		dc.b	"none",255

RAM_txt		dc.b	"RAM check:  ",255
LBank_txt	dc.b	"Lower bank: ",255
UBank_txt	dc.b	"Upper bank: ",255
kb_txt		dc.b	"KB, checksum = ",255
Erased_txt	dc.b	"erased",255
Blank_txt	dc.b	"blank",255
BJumper_txt	dc.b	"Warning: bank-select jumper may be open",255

Exit_txt	dc.b	"Press run/stop to exit",255
Retry_txt	dc.b	"Press R to retry"
retry_mod	dc.b	" after checking jumpers",255

Bank_txt	dc.b	"Upper or lower bank?",255
Erase_txt	dc.b	"Erasing ",255
Skipped_txt	dc.b	"skipped",255
Flash_txt	dc.b	"Flashing ",255

PowerOff_txt	dc.b	"Remove flashmode jumper and",255
	
de01_ROM	HEX	40 48 50 58
		HEX	c0 c8 d0 d8
		HEX	60 68 70 78
		HEX	e0 e8 f0 f8

RomSzBCD	HEX	04 08 16 32 64
RomSecSz	HEX	08 10 20 40 80		; #sectors in image

;	------------------------------------------------------------------

	MAC	module

 if	{1}=4
	dc.b	0
 endif
 if	{1}=8
	dc.b	1
 endif
 if	{1}=16
	dc.b	2
 endif
 if	{1}=32
	dc.b	3
 endif
 if	{1}=64
	dc.b	4
 endif

	dc.b	<[{2}],>[{2}]
	dc.b	{3},255
	ENDM

KnownROMs

;	module	size,chksum,"<max 28 chars>"

;	oldcarts - 32 KB		11

	module	32,$dcab,"Action Replay 4.2 PAL"
	module	32,$27db,"Action Replay 5.0 NTSC"
	module	32,$e805,"Action Replay 5.0 Triumwyrat"
	module	32,$5a5e,"Action Replay 6 Triad"
	module	32,$de4a,"Action Replay 6.0 PAL German"
	module	32,$2ba1,"Action Replay 6.0 PAL"
	module	32,$9e68,"Action Replay 7.0 PAL Derko"
	module	32,$2464,"Action Replay 7.3 PAL Atrax"
	module	32,$9548,"Action Replay 7.5 hacked"
	module	32,$51a2,"Atomic Power 1.1"
	module	32,$66d1,"Nordic Power 1.0 German"

;	Http-Load			3

	module	32,$96b3,"Http-Load 2004-07-07"
	module	32,$0008,"Http-Load 2004-07-14"
	module	32,$8efe,"Http-Load 2004-09-19"

;	The Final Replay		6

	module	32,$5958,"The Final Replay 0.1"
	module	32,$7725,"The Final Replay 0.2"
	module	32,$2033,"The Final Replay 0.3"
	module	32,$302a,"The Final Replay 0.4"
	module	32,$4cc2,"The Final Replay 0.5"
	module	32,$db38,"The Final Replay 0.6"

;	RetroReplay 32 KB		6

	module	32,$c555,"Retro Replay beta 2 NTSC"
	module	32,$9316,"Retro Replay beta 2 PAL"
	module	32,$515b,"Retro Replay beta 3 NTSC"
	module	32,$f24d,"Retro Replay beta 3 PAL"
	module	32,$bb34,"Retro Replay beta 3.1 NTSC"
	module	32,$1822,"Retro Replay beta 3.1 PAL"

;	RetroReplay 64 KB		8

	module	64,$f37d,"Retro Replay beta 3 NTSC"
	module	64,$c217,"Retro Replay beta 3 PAL"
	module	64,$20d4,"Retro Replay beta 3.1 NTSC"
	module	64,$11be,"Retro Replay beta 3.1 PAL"
	module	64,$1f41,"Retro Replay 3.8a NTSC"
	module	64,$e884,"Retro Replay 3.8a PAL"
	module	64,$54cc,"Retro Replay 3.8b PAL"
	module	64,$2fea,"Retro Replay 3.8p PAL"

;					34 AR/RR ROMs


;	4 KB game/utility images	13

	module	4,$4fd0,"Alien Sidestep"
	module	4,$2296,"Bug Crusher"
	module	4,$95d8,"Dot Gobbler"
	module	4,$91cd,"Close Encounters"
	module	4,$a47a,"Dot Gobbler v2"
	module	4,$c40b,"Gridrunner"
	module	4,$2d2b,"Laser Cycles"
	module	4,$6237,"Marios Brewery"
	module	4,$259b,"Space Action"
	module	4,$b9d3,"Space Ricoshay"
	module	4,$28e2,"Tank Wars"
	module	4,$2acb,"Hesmon64"
	module	4,$6138,"Retro Ball"

;	8 KB game/utility images	66

	module	8,$99f5,"Alpha Build"
	module	8,$11c4,"Astroblitz"
	module	8,$f32e,"Attack of the Mutant Camels"
	module	8,$ceaa,"Centipede"
	module	8,$0bcb,"Cosmic Life"
	module	8,$d58e,"Dance Fantasy"
	module	8,$962f,"Facemaker"
	module	8,$9489,"Fats Pool"
	module	8,$76bb,"Fraction Fever"
	module	8,$84f5,"Hop Along Counting"
	module	8,$db3d,"Jawbreaker"
	module	8,$feed,"Jukebox"
	module	8,$23e3,"Kids on Keys"
	module	8,$b0ac,"Kindercomp"
	module	8,$98b8,"Lazerzone"
	module	8,$5e3a,"Mazeman"
	module	8,$2b5f,"Memory Manor"
	module	8,$3234,"Mr Cool"
	module	8,$53e6,"Mr TNT"
	module	8,$1dda,"Number Tumbler"
	module	8,$202f,"Pacman"
	module	8,$cf0d,"Pipes"
	module	8,$ab23,"Save New York"
	module	8,$5fda,"Sea Speller"
	module	8,$5bd1,"Songmaker"
	module	8,$e344,"Stix"
	module	8,$52f6,"The Pit"
	module	8,$c174,"Threshold"
	module	8,$34c1,"Trashman"
	module	8,$a277,"Turbo Mazeman"
	module	8,$3d29,"Tylers Dungeons"
	module	8,$92b0,"Up for Grabs"
	module	8,$dd10,"Calc Result"
	module	8,$8cd8,"Coccinelle"
	module	8,$b307,"Financial Advisor"
	module	8,$c998,"Frogmaster"
	module	8,$cb50,"Moondust"
	module	8,$06a2,"Number Nabber"
	module	8,$b2d3,"Omega Race"
	module	8,$b62a,"Pitfall"
	module	8,$7255,"Spitball"
	module	8,$45c2,"Starpost"
	module	8,$3f01,"Super Expander"
	module	8,$38ac,"Super Smash"

;	Ultimax mode ROMs

	module	8,$2ca5,"Avenger"
	module	8,$8a31,"Avenger Max"
	module	8,$d465,"Clowns"
	module	8,$fc87,"Clowns Max"
	module	8,$d065,"Jupiter Lander"
	module	8,$8eb2,"Kickman"
	module	8,$2245,"Le Mans"
	module	8,$fba2,"Mini Basic"
	module	8,$7aad,"Mole Attack"
	module	8,$7c3d,"Money Wars"
	module	8,$51e5,"Music Composer"
	module	8,$1457,"Music Machine"
	module	8,$f6b8,"Omega Race Max"
	module	8,$0ffa,"Radar Rat Race"
	module	8,$11cf,"Radar Rat Race Max"
	module	8,$5519,"Road Race"
	module	8,$1432,"Sea Wolf"
	module	8,$618a,"Speedbingo Math"
	module	8,$d4f8,"Super Alien"
	module	8,$518e,"Visible Solar System"
	module	8,$2e60,"Wizard of Wor Max"
	module	8,$1f3d,"Wizard of Wor Max2"

;					79 game ROMs

	dc.b	-1,"unknown",255
Unknown_txt = *-8					; !! part of ROM table !!

;	------------------------------------------------------------------

;	Patch data for configurable ROMs

PatchPtr
	dc.w	p_3e00_10,p_7e00_10
	dc.w	p_3e00_10
	dc.w	p_1000_14,p_7e00_0a
	dc.w	p_1000_14,p_7e00_0a
	dc.w	p_1000_14

	dc.w	p_21c3_16
	dc.w	p_2263_16

p_3e00_10
p_7e00_10	HEX	9e00 0f 00

p_1000_14	HEX	9000 13 00
p_7e00_0a	HEX	9e00 09 00

p_21c3_16	HEX	81c3 15 00
p_2263_16	HEX	8263 15 00

;	screen layout

; 0	....Retro Replay Flasher Plugin V1.0....
; 1
; 2	Input file: 32 KB, checksum = 2BA1
; 3	            Action Replay 6.0 PAL      
; 4
; 5	Retro Replay RAM found
; 6	Flashmode jumper set
; 7	Flash manufacturer & device: 01 20
; 8	AMD 29F010
; 9	Protected sectors: none
;10
;11	RAM check:  ok
;12	Lower bank: 64 KB, checksum = 2FEA
;13	            Retro Replay 3.8p PAL
;14	Upper bank: 32 KB, checksum = D3B8
;15	            The Final Replay 0.6
;16*	Warning: bank-select jumper may be open
;17
;18	Press run/stop to exit
;19*	Press R to retry after checking jumpers
;20
;21	Upper or lower bank? x
;...
;18	Upper or lower bank? x
;19	Erasing ok
;20	Flashing ok
;21
;22	Press R to retry
;23	Remove flashmode jumper and
;24	press run/stop to exit